"""Defines labels and messages used in the context of an optimization service.

Copyright (c) 2023 The INVRS-IO authors.
"""

VALUE = "value"
GRAD = "grad"
PARAMS = "params"
STATE_TOKEN = "state_token"
MESSAGE = "message"

OPT_CONFIG = "opt_config"
ALGORITHM = "algorithm"
HPARAMS = "hparams"
DATA = "data"

ROUTE_INIT = "init"
ROUTE_UPDATE = "update"
ROUTE_PARAMS = "params"

MESSAGE_STATE_NOT_KNOWN = "State token {} was not recognized."
MESSAGE_STATE_NOT_READY = "State for token {} is not ready."
MESSAGE_STATE_NOT_VALID = "State for token {} is not valid."



"""Defines basic client optimizers for use with an optimization service.

Copyright (c) 2023 The INVRS-IO authors.
"""

import json
import time
from typing import Any, Dict, Optional

import requests
from totypes import json_utils

from invrs_opt.experimental import labels
from invrs_opt.optimizers import base

PyTree = Any
StateToken = str

SESSION = None
SERVER_ADDRESS = None


def login(server_address: str) -> None:
    """Set the global server address and create a requests session."""
    global SESSION
    global SERVER_ADDRESS
    SESSION = requests.Session()
    SERVER_ADDRESS = server_address


def optimizer_client(
    algorithm: str,
    hparams: Dict[str, Any],
    server_address: Optional[str],
    session: Optional[requests.Session],
) -> base.Optimizer:
    """Generic optimizer class."""

    if server_address is None:
        if SERVER_ADDRESS is None:
            raise ValueError(
                "Argument `server_address` and the global `SERVER_ADDRESS` cannot "
                "both be `None`. Use the `login` method to set the global, or "
                "explicitly provide a value."
            )
    if session is None:
        if SESSION is None:
            raise ValueError(
                "Argument `session` and the global `SESSION` cannot "
                "both be `None`. Use the `login` method to set the global, or "
                "explicitly provide a value."
            )
        session = SESSION

    opt_config = {
        labels.ALGORITHM: algorithm,
        labels.HPARAMS: hparams,
    }

    def init_fn(params: PyTree) -> StateToken:
        """Handles 'init' requests."""
        serialized_data = json_utils.json_from_pytree(
            dict(opt_config=opt_config, data={"params": params})
        )
        post_response = session.post(
            f"{SERVER_ADDRESS}/{labels.ROUTE_INIT}/", data=serialized_data
        )

        if not post_response.status_code == 200:
            raise requests.RequestException(post_response.text)
        response = json.loads(post_response.text)
        new_state_token: str = response[labels.STATE_TOKEN]
        return new_state_token

    def update_fn(
        *,
        grad: PyTree,
        value: float,
        params: PyTree,
        state: StateToken,
    ) -> StateToken:
        """Handles 'update' requests."""
        state_token = state
        del state
        serialized_data = json_utils.json_from_pytree(
            {
                labels.OPT_CONFIG: opt_config,
                labels.DATA: {
                    labels.PARAMS: params,
                    labels.VALUE: value,
                    labels.GRAD: grad,
                    labels.STATE_TOKEN: state_token,
                },
            }
        )
        post_response = session.post(
            f"{SERVER_ADDRESS}/{labels.ROUTE_UPDATE}/{state_token}/",
            data=serialized_data,
        )

        if not post_response.status_code == 200:
            raise requests.RequestException(post_response.text)
        response = json.loads(post_response.text)
        new_state_token: str = response[labels.STATE_TOKEN]
        return new_state_token

    def params_fn(
        state: StateToken,
        timeout: float = 60.0,
        poll_interval: float = 0.1,
    ) -> PyTree:
        """Handles 'params' requests."""
        state_token = state
        del state
        assert timeout >= poll_interval
        start_time = time.time()
        while time.time() < start_time + timeout:
            get_response = session.get(
                f"{SERVER_ADDRESS}/{labels.ROUTE_PARAMS}/{state_token}"
            )
            if get_response.status_code == 200:
                break
            elif get_response.status_code == 404 and get_response.text.endswith(
                labels.MESSAGE_STATE_NOT_READY.format(state_token)
            ):
                time.sleep(poll_interval)
            else:
                raise requests.RequestException(get_response.text)

        if not get_response.status_code == 200:
            raise requests.Timeout("Timed out while waiting for params.")
        response = json.loads(get_response.text)
        return json_utils.pytree_from_json(response[labels.PARAMS])

    return base.Optimizer(init=init_fn, update=update_fn, params=params_fn)


# -----------------------------------------------------------------------------
# Specific optimizers implemented here.
# -----------------------------------------------------------------------------


def lbfgsb(maxcor: int = 20, line_search_max_steps: int = 100) -> base.Optimizer:
    """Optimizer implementing the L-BFGS-B scheme."""
    hparams = {
        "maxcor": maxcor,
        "line_search_max_steps": line_search_max_steps,
    }
    return optimizer_client(
        algorithm="lbfgsb", hparams=hparams, server_address=None, session=None
    )



"""Defines the direct pixel parameterization for density arrays.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses

import jax.numpy as jnp
from jax import tree_util
from totypes import json_utils, types

from invrs_opt.parameterization import base


@dataclasses.dataclass
class PixelParams(base.ParameterizedDensity2DArray):
    latents: "PixelLatents"
    metadata: None = None


@dataclasses.dataclass
class PixelLatents(base.LatentsBase):
    """Stores latent parameters for the direct pixel parameterization."""

    density: types.Density2DArray


tree_util.register_dataclass(
    PixelParams,
    data_fields=["latents"],
    meta_fields=[],
)
tree_util.register_dataclass(
    PixelLatents,
    data_fields=["density"],
    meta_fields=[],
)
json_utils.register_custom_type(PixelParams)
json_utils.register_custom_type(PixelLatents)


def pixel() -> base.Density2DParameterization:
    """Return the direct pixel parameterization."""

    def from_density_fn(density: types.Density2DArray) -> PixelParams:
        return PixelParams(latents=PixelLatents(density=density))

    def to_density_fn(params: PixelParams) -> types.Density2DArray:
        return params.latents.density

    def constraints_fn(params: PixelParams) -> jnp.ndarray:
        del params
        return jnp.asarray(0.0)

    def update_fn(params: PixelParams, step: int) -> PixelParams:
        del step
        return params

    return base.Density2DParameterization(
        from_density=from_density_fn,
        to_density=to_density_fn,
        constraints=constraints_fn,
        update=update_fn,
    )



"""Defines filter-and-project density parameterization.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses

import jax.numpy as jnp
from jax import tree_util
from totypes import json_utils, types

from invrs_opt.parameterization import base, transforms


@dataclasses.dataclass
class FilterProjectParams(base.ParameterizedDensity2DArray):
    """Stores parameters for the filter-project parameterization."""

    latents: "FilterProjectLatents"
    metadata: "FilterProjectMetadata"


@dataclasses.dataclass
class FilterProjectLatents(base.LatentsBase):
    """Stores latent parameters for the filter-project parameterization.

    Attributes:s
        latent_density: The latent variable from which the density is obtained.
    """

    latent_density: types.Density2DArray


@dataclasses.dataclass
class FilterProjectMetadata(base.MetadataBase):
    """Stores metadata for the filter-project parameterization.

    Attributes:
        beta: Determines the sharpness of the thresholding operation.
    """

    beta: float


tree_util.register_dataclass(
    FilterProjectParams,
    data_fields=["latents", "metadata"],
    meta_fields=[],
)
tree_util.register_dataclass(
    FilterProjectLatents,
    data_fields=["latent_density"],
    meta_fields=[],
)
tree_util.register_dataclass(
    FilterProjectMetadata,
    data_fields=[],
    meta_fields=["beta"],
)
json_utils.register_custom_type(FilterProjectParams)
json_utils.register_custom_type(FilterProjectLatents)
json_utils.register_custom_type(FilterProjectMetadata)


def filter_project(beta: float) -> base.Density2DParameterization:
    """Defines a filter-project parameterization for density arrays.

    The `DensityArray2D` is represented as latent density array that is transformed by,

        transformed = tanh(beta * conv(density.array, gaussian_kernel)) / tanh(beta)

    where the kernel has a full-width at half-maximum determined by the minimum width
    and spacing parameters of the `DensityArray2D`.

    When the density lower and upper bounds are -1 and +1, this basic expression is
    Where the bounds differ, the density is scaled before the transform is applied, and
    then unscaled afterwards.

    Args:
        beta: Determines the sharpness of the thresholding operation.

    Returns:
        The `Density2DParameterization`.
    """

    def from_density_fn(density: types.Density2DArray) -> FilterProjectParams:
        """Return latent parameters for the given `density`."""
        array = transforms.normalized_array_from_density(density)
        array = jnp.clip(array, -1, 1)
        array *= jnp.tanh(beta)
        latent_array = jnp.arctanh(array) / beta
        latent_array = transforms.rescale_array_for_density(latent_array, density)
        latent_density = density = dataclasses.replace(density, array=latent_array)
        return FilterProjectParams(
            latents=FilterProjectLatents(latent_density=latent_density),
            metadata=FilterProjectMetadata(beta=beta),
        )

    def to_density_fn(params: FilterProjectParams) -> types.Density2DArray:
        """Return a density from the latent parameters."""
        latent_density = params.latents.latent_density
        beta = params.metadata.beta

        transformed = types.symmetrize_density(latent_density)
        transformed = transforms.density_gaussian_filter_and_tanh(transformed, beta)
        # Scale to ensure that the full valid range of the density array is reachable.
        mid_value = (transformed.lower_bound + transformed.upper_bound) / 2
        transformed = tree_util.tree_map(
            lambda array: mid_value + (array - mid_value) / jnp.tanh(beta), transformed
        )
        return transforms.apply_fixed_pixels(transformed)

    def constraints_fn(params: FilterProjectParams) -> jnp.ndarray:
        """Computes constraints associated with the params."""
        del params
        return jnp.asarray(0.0)

    def update_fn(params: FilterProjectParams, step: int) -> FilterProjectParams:
        """Perform updates to `params` required for the given `step`."""
        del step
        return params

    return base.Density2DParameterization(
        to_density=to_density_fn,
        from_density=from_density_fn,
        constraints=constraints_fn,
        update=update_fn,
    )



"""Defines transforms for custom types and bare jax arrays.

Copyright (c) 2023 The INVRS-IO authors.
"""

from typing import Tuple, Union

import jax
import jax.numpy as jnp
from jax import tree_util
from totypes import types

PadMode = Union[str, Tuple[str, str]]

GAUSSIAN_FWHM_SIZE_MULTIPLE: float = 3.0


def density_gaussian_filter_and_tanh(
    density: types.Density2DArray,
    beta: float,
    fwhm_size_multiple: float = GAUSSIAN_FWHM_SIZE_MULTIPLE,
) -> types.Density2DArray:
    """Filters the density using a Gaussian kernel and applies a tanh nonlinearity.

    The Gaussian full width at half maximum is the average of the minimum width and
    minimum spacing for the density.

    Args:
        density: The density to be transformed.
        beta: Scalar determining the strength of the tanh nonlinearity.
        fwhm_size_multiple: Determines the size of the Gaussian kernel; if the full
            width is `fwhm`, then the kernel size is `ceil(fwhm_size_multiple * fwhm)`.

    Returns:
        The transformed `density`.
    """
    array = normalized_array_from_density(density)
    pad_mode = _pad_mode_for_density(density)

    batch_shape = array.shape[:-2]
    array_flat_batch = array.reshape((-1, 1) + array.shape[-2:])

    kernel = _gaussian_kernel(
        (density.minimum_width + density.minimum_spacing) / 2,
        fwhm_size_multiple=fwhm_size_multiple,
    )
    kernel /= jnp.sum(kernel)
    kernel = kernel[jnp.newaxis, jnp.newaxis, :, :]  # OIHW format

    pad_width = pad_width_for_kernel_shape(kernel.shape)
    transformed = pad2d(array_flat_batch, pad_width, pad_mode=pad_mode)
    transformed = conv(transformed, kernel, padding="VALID")
    transformed = jnp.tanh(beta * transformed)
    transformed = transformed[..., 0, :, :]  # Remove the dummy channel axis.
    transformed = transformed.reshape((batch_shape) + array.shape[-2:])
    transformed = rescale_array_for_density(transformed, density)

    treedef = jax.tree_util.tree_structure(density)
    transformed_density: types.Density2DArray = jax.tree_util.tree_unflatten(
        treedef, (transformed,)
    )
    return transformed_density


def _gaussian_kernel(fwhm: float, fwhm_size_multiple: float) -> jnp.ndarray:
    """Returns a Gaussian kernel with the specified full-width at half-maximum."""
    with jax.ensure_compile_time_eval():
        kernel_size = max(1, int(jnp.ceil(fwhm * fwhm_size_multiple)))
    # Ensure the kernel size is odd, so that there is always a central pixel which will
    # contain the peak value of the Gaussian.
    kernel_size += (kernel_size + 1) % 2
    d = jnp.arange(0.5, kernel_size) - kernel_size / 2
    x = d[:, jnp.newaxis]
    y = d[jnp.newaxis, :]
    sigma = fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))
    return jnp.exp(-(x**2 + y**2) / (2 * sigma**2))


def normalized_array_from_density(density: types.Density2DArray) -> jnp.ndarray:
    """Returns an array with values scaled to the range `(-1, 1)`."""
    value_mid = (density.upper_bound + density.lower_bound) / 2
    value_range = density.upper_bound - density.lower_bound
    return jnp.asarray(2 * (density.array - value_mid) / value_range)


def rescale_array_for_density(
    array: jnp.ndarray,
    density: types.Density2DArray,
) -> jnp.ndarray:
    """Rescales an array for the bounds defined by `density`."""
    value_mid = (density.upper_bound + density.lower_bound) / 2
    value_range = density.upper_bound - density.lower_bound
    return array / 2 * value_range + value_mid


def apply_fixed_pixels(density: types.Density2DArray) -> types.Density2DArray:
    """Set fixed pixels with their required values."""
    fixed_solid = density.fixed_solid
    fixed_void = density.fixed_void
    (array,), treedef = tree_util.tree_flatten(density)
    if fixed_solid is not None:
        array = jnp.where(fixed_solid, density.upper_bound, array)
    if fixed_void is not None:
        array = jnp.where(fixed_void, density.lower_bound, array)
    transformed_density: types.Density2DArray = tree_util.tree_unflatten(
        treedef, (array,)
    )
    return transformed_density


def conv(x: jnp.ndarray, kernel: jnp.ndarray, padding: str) -> jnp.ndarray:
    """Convolves `x` with `kernel`."""
    assert x.ndim == 4
    assert kernel.ndim == 4
    return jax.lax.conv_general_dilated(
        lhs=x,
        rhs=kernel,
        padding=padding,
        dimension_numbers=("NCHW", "OIHW", "NCHW"),
        window_strides=(1, 1),
    )


def pad_width_for_kernel_shape(
    kernel_shape: Tuple[int, ...],
) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """Return the pad width for the trailing dimensions of the kernel."""
    kernel_height, kernel_width = kernel_shape[-2:]
    return (
        (kernel_height // 2, kernel_height // 2 - (kernel_height + 1) % 2),
        (kernel_width // 2, kernel_width // 2 - (kernel_width + 1) % 2),
    )


def pad2d(
    x: jnp.ndarray,
    pad_width: Tuple[Tuple[int, int], Tuple[int, int]],
    pad_mode: PadMode,
) -> jnp.ndarray:
    """Pad the trailing two axes of `x` as specified."""
    leading_pad_width = ((0, 0),) * (x.ndim - 2)
    if isinstance(pad_mode, str):
        return jnp.pad(x, leading_pad_width + pad_width, mode=pad_mode)

    pad_width_i, pad_width_j = pad_width
    pad_mode_i, pad_mode_j = pad_mode
    x_padded_i = jnp.pad(
        x,
        pad_width=leading_pad_width + (pad_width_i, (0, 0)),
        mode=pad_mode_i,
    )
    return jnp.pad(
        x_padded_i,
        pad_width=leading_pad_width + ((0, 0), pad_width_j),
        mode=pad_mode_j,
    )


def _pad_mode_for_density(density: types.Density2DArray) -> Union[str, Tuple[str, str]]:
    """Return the pad mode implied by the `periodic` attribute of `density`."""
    if density.periodic == (True, True):
        return "wrap"
    if density.periodic == (False, False):
        return "edge"
    return (
        "wrap" if density.periodic[0] else "edge",
        "wrap" if density.periodic[1] else "edge",
    )


def resample(
    x: jnp.ndarray,
    shape: Tuple[int, ...],
    method: jax.image.ResizeMethod = jax.image.ResizeMethod.LINEAR,
) -> jnp.ndarray:
    """Resamples `x` to have the specified `shape`.

    The algorithm first upsamples `x` so that the pixels in the output image are
    comprised of an integer number of pixels in the upsampled `x`, and then
    performs box downsampling.

    Args:
        x: The array to be resampled.
        shape: The shape of the output array.
        method: The method used to resize `x` prior to box downsampling.

    Returns:
        The resampled array.
    """
    if x.ndim != len(shape):
        raise ValueError(
            f"`shape` must have length matching number of dimensions in `x`, "
            f"but got {shape} when `x` had shape {x.shape}."
        )

    with jax.ensure_compile_time_eval():
        factor = [int(jnp.ceil(dx / d)) for dx, d in zip(x.shape, shape)]
        upsampled_shape = tuple([d * f for d, f in zip(shape, factor)])

    x_upsampled = jax.image.resize(
        image=x,
        shape=upsampled_shape,
        method=method,
    )

    return box_downsample(x_upsampled, shape)


def box_downsample(x: jnp.ndarray, shape: Tuple[int, ...]) -> jnp.ndarray:
    """Downsamples `x` to a coarser resolution array using box downsampling.

    Box downsampling forms nonoverlapping windows and simply averages the
    pixels within each window. For example, downsampling `(0, 1, 2, 3, 4, 5)`
    with a factor of `2` yields `(0.5, 2.5, 4.5)`.

    Args:
        x: The array to be downsampled.
        shape: The shape of the output array; each axis dimension must evenly
            divide the corresponding axis dimension in `x`.

    Returns:
        The output array with shape `shape`.
    """
    if x.ndim != len(shape) or any([(d % s) != 0 for d, s in zip(x.shape, shape)]):
        raise ValueError(
            f"Each axis of `shape` must evenly divide the corresponding axis "
            f"dimension in `x`, but got shape {shape} when `x` has shape "
            f"{x.shape}."
        )
    shape = sum([(s, d // s) for d, s in zip(x.shape, shape)], ())
    axes = list(range(1, 2 * x.ndim, 2))
    x = x.reshape(shape)
    return jnp.mean(x, axis=axes)
